-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore DATA_RAM.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- The data/return stack RAM implemented in such a way
-- that it should be recognised as asynchronous 
-- RAM cells by the synthesiser.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;


ENTITY data_ram IS
PORT (dram : INOUT dram_port); -- The whole RAM interface is contained in RECORD dram_port
END data_ram;

ARCHITECTURE asynchronous_ram OF data_ram IS

TYPE ram_data IS ARRAY (INTEGER RANGE 0 TO 2**data_addr_width-1) OF data_bus;

SIGNAL clk   : std_logic;
SIGNAL write : std_logic;  
SIGNAL addr  : data_addr;
SIGNAL din   : data_bus;
SIGNAL ram   : ram_data;

BEGIN
-----------------------------------------------------------------------
-- ram_model: Input data are taken in iff clk edge & WE = 1
-----------------------------------------------------------------------

clk   <= dram.clk;
write <= dram.write;
addr  <= dram.addr;
din   <= dram.din;

data_ram: PROCESS(clk)
BEGIN
  IF  rising_edge(clk)  THEN
     IF  write='1'  THEN
         ram(to_int('0' & addr)) <= din;
     END IF;
  END IF;
END PROCESS data_ram;

dram.dout <= ram(to_int('0' & addr));

END asynchronous_ram;
