\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/License/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore MICROCORE.F 1.10
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\
\ Introduction  see: information/uCore_Introduction.pdf
\ Tools         see: information/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: information/uCore_License.pdf
\ Conventions   see: information/uCore_Conventions.pdf
\ Contributors  see: information/uCore_Contributors.pdf
\ Philosophy    see: information/uCore_Philosophy.pdf
\
\ MicroCore Forth Cross-compiler
\ This code loads on the public domain Win32Forth http://www.forth.org/
\ Known Limitations:
\ V_1.10 can at most handle a data width of 31 bits.
\
Forth definitions decimal

: .version  ( -- )  ." 1.10 " ;
: .date     ( -- )  ." 15-Jan-2003 " ;

: 2**  ( n -- 2**n )   1 swap 0 ?DO  2*  LOOP ;

#prog_addr_width 2**             Constant #maxprog  
#data 2**                        Constant #prog
#prog 1-                         Constant #progmask

#code 2** 1-                     Constant #codemask

#data #code 1- /mod swap 0= 1+ + Constant #nibbles  \ nibbles needed to represent any number

\
\ Some System word (re)definitions for a more sympathetic environment
\

False TO sys-warning?   \ switch off Win32For-"blabla"

$1B Constant #esc

: rdrop  ( -- )             postpone r> postpone drop ; immediate

: shift  ( n1 quan -- n2 )  dup 0< IF  abs rshift  EXIT THEN  lshift ;

: ndrop  ( n1 .. nn n -- )   0 DO  drop  LOOP ;

\ The following two words SOURCE> and >SOURCE are needed because the implementation
\ of >IN is not according to the standard. Used in order to re-interpret the source code.
\ Source code re-interpretation is used to handle the variable address field
\ size of microcore branches

: source>  ( -- sca )   source-position $10 shift   >in @ $FFFF and or ;

: >source  ( sca -- )
   dup $-10 shift
   dup source-position - IF  0 source-id FSEEK-FILE drop refill  THEN
   drop  $FFFF and >in ! ;

\ Create restore ] r> r> ! UNNEST [
\ : push  ( addr -- )   r> swap dup @ >r >r restore rel>abs >r >r ;
\ this PUSH seems to corrupt the header dictionary space table??? mistery

: comp?  ( -- f )   State @ 0<> ;
: exec?  ( -- f )   State @ 0= ;

: ?pairs  ( n1 n2 -- )   - abort" unstructured" ;                       \ for checking conditional constructs

: :       ( -- 0 )   : 0 ;                                              \ this allows to create an "intelligent" ; later on
                                                                        \ for MACRO definitions to pick the proper
: ;       ( 0 -- )   0 ?pairs postpone ; [ 0 ?pairs ] ; immediate       \ ; behaviour

: 0=EXIT  ( f -- )   postpone 0= postpone ?EXIT ; immediate

: case?   ( n1 n2 -- n1 ff | tf )   over = dup IF  nip  THEN ;          \ the most primitive case operator
\
\ accessing the targets program memory
\
Create Memory   #prog #maxprog min cells allot

Variable Tcp                        \ TargetCodePointer
Variable Tdp                        \ TargetDataPointer
Variable Sequential                 \ number of past opcodes available for peep-hole optimization

Variable Macro      Macro off       \ holds the current macro tag to be included in code compilation. 0 => no macro
Variable Macros     Macros off      \ holds the last assigned macro number
Variable Macro-link Macro-link off  \ linked list of all macros

$FF000000 Constant #macromask  \ macro# is embedded in opcode field

: >memory  ( caddr -- addr )   cells Memory + ;
: t!       ( n caddr -- )      >memory ! ;
: t@       ( caddr -- n )      >memory @ ;
: opcode@  ( caddr -- op )     t@ #codemask and ;
: macro@   ( caddr -- # )      t@ $-18 shift ;
: macro!   ( n caddr -- )      swap Macro @ $18 shift or swap t! ;
: there    ( -- addr )         Tcp @ ;
: tallot   ( n -- )            dup Sequential +!   Tcp +! ;
: t,       ( n -- )
   dup [ #codemask not ] Literal and abort" not an opcode"
   there macro!  1 Tcp +!  1 Sequential +!
;
: .addr   ( caddr -- )   Base @ >r  hex  
   0 <#  [ #data 3 + 4 / ] Literal 0 DO  #  LOOP  #> type
   r> Base !
;
: .opcode ( opcode -- )  Base @ >r  hex  
   0 <#  [ #code 3 + 4 / ] Literal 0 DO  #  LOOP  #> type space
   r> Base !
;
\
\ peep-hole optimizer
\
: |        ( -- )              Sequential off ;             \ break peep-hole optimizer

: prev@    ( -- opcode | 0 )   Sequential @     IF  there 1 - t@  EXIT THEN  false ;

: pprev@   ( -- opcode | 0 )   Sequential @ 1 > IF  there 2 - t@  EXIT THEN  false ;

: Match  ( opcode mask -- )    Create  , ,
Does>  ( opcode -- f )  dup @ >r   cell+ @ xor r> and 0= ;

\
\ instruction fields as defined by the hardware architecture
\
$80 Constant #literal
$7F Constant #litmask
$60 Constant #op
$18 Constant #stack
$07 Constant #group

$00 Constant #bra
$20 Constant #alu
$40 Constant #mem
$60 Constant #usr

$00 Constant #none
$08 Constant #pop
$10 Constant #push
$18 Constant #both

\ binary Operators
  0 Constant #add
  1 Constant #adc
  2 Constant #sub
  3 Constant #sbc
  4 Constant #and
  5 Constant #or
  6 Constant #xor
  7 Constant #nos

\ unary operators
  0 Constant #not
  1 Constant #sl
  2 Constant #asr
  3 Constant #lsr
  4 Constant #ror
  5 Constant #rol
  6 Constant #zequ
  7 Constant #cc

\ conditions
  0 Constant #never
  1 Constant #zero
  2 Constant #nsign
  3 Constant #ncarry
  4 Constant #pause
  5 Constant #int
  6 Constant #nzero
  7 Constant #always

\ registers
  0 Constant #status
  1 Constant #tor
  2 Constant #rstack
  3 Constant #local
  4 Constant #rsp
  5 Constant #dsp
  6 Constant #task
  7 Constant #flags

\
\ Opcode-field compilers and Opcode compiler
\
: Type:  ( <name> n -- )
   Create , 
Does> ( u1 u2 -- opcode )
   3 depth > abort" not enough fields specified"
   @ over #stack not and abort" not a stack field"  or
   over #group not and abort" not a group field"    or
;      

: dont   ( -- )  True abort" can not be executed" ;

: Field:  ( <name> n -- )  Constant ;

: Group:  ( <name> n -- )  Constant ;

Variable Operators  Operators off          \ Linked list of all operators, used by the disassembler

: Op:  ( n <name> -- )
   Create ,   \ here the instruction code is layed down which will be compiled into the target memory
        ' ,   \ here the cfa of the word is layed down which will be executed during interpretation in the target
   here  Operators @ , Operators !
Does> ( -- )  exec? IF  cell+ @ execute  EXIT THEN  @ t, ;

Create nomacro  ] Macro off  UNNEST [
: (macro  ( -- )  \ handles in-line macro# and link-field: | (macro | macro# | link |
   r>   nomacro rel>abs >r
   dup cell+ cell+ >r   abs>rel @ Macro !
;
: Macro:  ( <name> -- 0 )
   :   1 Macros +!   postpone (macro   Macros @ ,
   here Macro-link @ , Macro-link !
;
\
\ The MicroCore "Assembler"
\
Vocabulary Target                \ Vocabulary for all Target definitions

: H  ( -- )  Forth  ; immediate  \ "Host" - short-hand definition to switch context also within a definition
: T  ( -- )  Target ; immediate

Target definitions

Vocabulary Assembler

: {  ( -- )   Assembler ; immediate     \ The braces allow to "delve" into the assembler during target compilation

: Asm  ( -- )   T Assembler H ;

Forth definitions

: Asm  ( -- )   T Assembler H ; immediate

Target Assembler definitions

: }  ( -- )   Target ; immediate        \ This switches back into "Forth"-mode

#alu Type: ALU
#bra Type: BRA
#mem Type: MEM
#usr Type: USR

#none Field: NONE
#push Field: PUSH
#pop  Field: POP
#both Field: BOTH

\ Binary operators
#add   Group: ADD
#sub   Group: SUB
#adc   Group: ADC
#sbc   Group: SBC
#and   Group: AND
#or    Group: OR
#xor   Group: XOR
#nos   Group: NOS

\ Unary operators
#not   Group: NOT
#sl    Group: SL
#asr   Group: ASR
#lsr   Group: LSR
#ror   Group: ROR
#rol   Group: ROL
#zequ  Group: ZEQU
#cc    Group: CC

\ Conditions
#never  Group: NEVER
#zero   Group: ZERO
#nsign  Group: NSIGN
#ncarry Group: NCARRY
#pause  Group: PAUSE
#int    Group: INT
#nzero  Group: NZERO
#always Group: ALWAYS

\ Register
#status  Group: STATUS
#tor     Group: TOR
#rstack  Group: RSTACK
#local   Group: LOCAL
#rsp     Group: RSP
#dsp     Group: DSP
#task    Group: TASK
#flags   Group: FLAGS

\ offsets for post-incremeting @ and ! instructions
 0 Group: 0
 1 Group: 1
 2 Group: 2
 3 Group: 3
 4 Group: -4
 5 Group: -3
 6 Group: -2
 7 Group: -1

\
\ Target literals and addresses are composed of literal-nibbles
\ preceeding opcodes which use them
\
Target definitions Assembler

NEVER  NONE BRA Op: nop     noop  ( -- )

Forth definitions

: /nibble  ( lit -- nib lit' )
   dup #litmask and #literal or   swap #literal /
;
: last_nibble?  ( nib lit -- f )
   0 case? IF  [ #literal u2/ ] Literal and 0=  EXIT THEN
   -1    = IF  [ #literal u2/ ] Literal and     EXIT THEN
   drop False
;
: >nibbles  ( lit -- nib_q .. nib_1 q )  \ convert literal into q nibbles
   0 >r  BEGIN  r> 1+ >r   /nibble  2dup last_nibble? UNTIL  drop r>
;
: nibbles  ( lit -- quan )  \ number of nibbles needed to represent lit
   0 >r  BEGIN  r> 1+ >r   /nibble  tuck last_nibble? UNTIL  drop r>
;
: *nibble  ( nib lit' -- lit )
   #literal *   swap #litmask and or
;
: nibbles>  ( nib_q .. nib_1 q -- lit )  \ convert q nibbles into literal
   dup 0= ABORT" no nibbles available"
   >r  #litmask and
   dup [ #literal u2/ ] Literal and IF  #literal negate or  THEN
   r> 1 ?DO  *nibble  LOOP
;
: nibbles!  ( nib_q .. nib_1 quan caddr -- )  swap bounds DO  I macro!  LOOP ;

: prevlit? ( -- f )   prev@ #literal and ;

: ?nop,  ( -- )      prevlit? 0=EXIT
   Macro @   Macro off  T nop H   Macro !
;
: nibbles,  ( nib_q .. nib_1 quan -- )
   ?nop, there over tallot nibbles!
;
: nibbles@  ( caddr -- nib_q .. nib_1 quan )
   dup  BEGIN  dup t@ #literal and WHILE  dup t@ >r  1+  REPEAT
   swap -   dup  BEGIN  ?dup WHILE  1-  r> -rot  REPEAT
;
: ?literal ( lit -- lit' )   \ range check and sign extension
   dup dup 0< IF  #progmask or 1+  ELSE  [ #progmask invert ] Literal and  THEN
   abort" number beyond data path width"
   dup [ #prog u2/ ] Literal and
   IF  [ #progmask invert ] Literal or  ELSE  #progmask and  THEN   \ sign extension
;
: lit, ( n -- )   \ compiles a literal into the minimum number of nibbles
   ?literal exec? ?EXIT  >nibbles nibbles,
;
: fixed,  ( n quan -- )      \ compiles a literal into quan nibbles.
   ?literal dup >r 0 DO  /nibble  LOOP  drop  r> nibbles,
;

DEBUGGING [IF]

: .lit  ( caddr -- )   nibbles@ nibbles> . ;

: tdump  ( caddr quan -- )
   cells bounds ?DO  cr I .addr ." :  "
                     I >memory $10 cells bounds DO  I @ .opcode  1 cells +LOOP
                $10 +LOOP ;

[THEN]
: prevsmall?  ( -- ff )
   prevlit? 0= abort" needs literal"   
   prev@ #litmask and [ #group u2/ invert ] Literal and   
   dup IF  [ #group u2/ 1+ ] Literal + #litmask and   THEN     \ either %0..00000 and %1..111-- is ok
;
\
\ Opcode definitions
\
Target definitions Assembler

RSTACK NONE MEM Op: >r           dont ( n -- )
RSTACK BOTH MEM Op: r>           dont ( -- n )
TOR    BOTH MEM Op: r@           dont ( -- n )
NEVER  PUSH BRA Op: dup          dup   ( n -- n n )
NEVER  POP  BRA Op: drop         drop  ( n -- )
NOS    NONE ALU Op: swap         swap  ( n1 n2 -- n2 n1 )
NOS    PUSH ALU Op: over         over  ( n1 n2 -- n1 n2 n1 )
ADD    POP  ALU Op: +            +     ( n1 n2 -- n3 )
ADD    PUSH ALU Op: 2dup_+       dont ( n1 n2 -- n1 n2 n3 )
SUB    NONE ALU Op: over_swap_-  dont ( n1 n2 -- n1 n3 )
ADC    POP  ALU Op: +c           dont
SUB    POP  ALU Op: -            -     ( n1 n2 -- n3 )
SBC    POP  ALU Op: -c           dont
AND    POP  ALU Op: and          and
OR     POP  ALU Op: or           or
XOR    POP  ALU Op: xor          xor
NOT    BOTH ALU Op: invert       invert
SL     BOTH ALU Op: 2*           2*
ASR    BOTH ALU Op: 2/           2/
LSR    BOTH ALU Op: u2/          u2/
ROR    BOTH ALU Op: ror          dont
ROL    BOTH ALU Op: rol          dont
ZEQU   BOTH ALU Op: 0=           0=
CC     BOTH ALU Op: cc           dont
ZERO   POP  BRA Op: z-branch     dont
ALWAYS POP  BRA Op: branch       dont
NCARRY POP  BRA Op: nc-branch    dont
NSIGN  POP  BRA Op: ns-branch    dont
NZERO  POP  BRA Op: nz-branch    dont
ALWAYS BOTH BRA Op: call         dont
ZERO   BOTH BRA Op: z-call       dont
NZERO  BOTH BRA Op: nz-call      dont
NCARRY BOTH BRA Op: nc-call      dont
NSIGN  BOTH BRA Op: ns-call      dont
PAUSE  BOTH BRA Op: pause        dont
ALWAYS NONE BRA Op: exit         dont
ZERO   NONE BRA Op: z-exit       dont
NZERO  NONE BRA Op: nz-exit      dont
NCARRY NONE BRA Op: nc-exit      dont
NSIGN  NONE BRA Op: ns-exit      dont
INT    NONE BRA Op: iret         dont
0      PUSH MEM Op: (@           dont
0      POP  MEM Op: (!           dont
LOCAL  BOTH MEM Op: ++l@         dont
LOCAL  NONE MEM Op: ++l!         dont
STATUS NONE MEM Op: status!      dont
STATUS BOTH MEM Op: status@      dont
FLAGS  NONE MEM Op: ie!          dont
FLAGS  BOTH MEM Op: flags@       dont
TASK   NONE MEM Op: ++u!         dont
TASK   BOTH MEM Op: ++u@         dont
DSP    NONE MEM Op: dsp!         dont
DSP    BOTH MEM Op: dsp@         dont
RSP    NONE MEM Op: rsp!         dont
RSP    BOTH MEM Op: rsp@         dont

Macro: @       ( addr -- n )    ?comp T (@ drop H ;
Macro: !       ( n addr -- )    ?comp T (! drop H ;
Macro: l@      ( rel -- n )     ?comp T ++l@ drop H ;
Macro: l!      ( n rel -- )     ?comp T ++l! drop H ;
Macro: u@      ( rel -- n )     ?comp T ++u@ drop H ;
Macro: u!      ( n rel -- )     ?comp T ++u! drop H ;
     : ++@     ( n -- )         ?comp prevsmall? IF  T + (@ H  EXIT THEN   
        prev@ #group and [ Asm  0 PUSH MEM H ] Literal or  -1 tallot t,
     ;
     : ++!     ( n -- )         ?comp prevsmall? IF  T + (! H  EXIT THEN   
        prev@ #group and [ Asm  0 POP MEM H ] Literal or  -1 tallot t,
     ;
Macro: rdrop   ( -- )                    ?comp T r>  drop H ;
Macro: 2drop   ( n n -- )                comp? IF  T drop drop       H EXIT THEN  2drop ;
Macro: 2dup    ( d -- d d )              comp? IF  T over over       H EXIT THEN  2dup ;
Macro: c=0     ( -- )                    ?comp  0 lit, T 2* drop H ;
Macro: c=1     ( -- )                    ?comp -1 lit, T 2* drop H ;
Macro: =       ( n1 n2 -- f )            comp? IF  T - 0=            H EXIT THEN  = ;
Macro: negate  ( n -- -n )               comp? IF  0 lit, T swap -   H EXIT THEN  negate ;
     : Literal ( -- n )                  lit, ;
Macro: 1+      ( n1 -- n2 )              comp? IF   1 lit, T +       H EXIT THEN  1+ ;
Macro: 1-      ( n1 -- n2 )              comp? IF  -1 lit, T +       H EXIT THEN  1- ;
Macro: True    ( -- tf )                -1 lit, ;
Macro: False   ( -- ff )                 0 lit, ;
Macro: rot     ( n1 n2 n3 -- n2 n3 n1 )  comp? IF  T >r swap r> swap H EXIT THEN  rot ;
Macro: -rot    ( n1 n2 n3 -- n3 n1 n2 )  comp? IF  T swap >r swap r> H EXIT THEN  -rot ;
Macro: nip     ( n1 n2 -- n2 )           comp? IF  T swap drop       H EXIT THEN  nip ;
Macro: tuck    ( n1 n2 -- n2 n1 n2 )     comp? IF  T swap over       H EXIT THEN  tuck ;
Macro: ?dup    ( n -- n n | 0 )          comp? IF  1 lit, T z-branch dup H | EXIT THEN  ?dup ;
Forth definitions  Macros @ Constant #?dup

\
\ branching
\ Due to relative addressing, the number of nibbles needed for a branch address
\ varies. At first, only one nibble is assumed to be sufficient to hold the offset.
\ When the compiler finds that more nibbles are needed, the source code is re-interpreted
\
Forth definitions

: >branch  ( caddr -- br.addr quan.nibbles )  \ length of branch lit field
   0  BEGIN  over t@ #literal and WHILE  1+ swap  1+ swap  REPEAT
;
: <mark    ( -- sca caddr )    ?nop, |  source> there ;

: <resolve ( sca caddr -- )  \ ProgramCounter points behind branch opcode
   nip   |   1 >r                 \ minimum length of branch
   BEGIN  dup there r@ + 1+ -     \ branch offset
          nibbles r@ >            \ does branch offset fit?
   WHILE  r> 1+ >r                \ try with offset one nibble longer
   REPEAT
   there r@ + 1+ -  r> fixed,
;
: >mark    ( quan.nibbles -- sca caddr )
   ?nop, |  source> there rot 0 ?DO  #literal t,  LOOP
;
: >resolve ( sca caddr -- quan.nibbles.needed | 0 )
   |   dup >r   >branch >r          \ R: offset width allocated
   there swap 1+ - >nibbles         \ branch offset
   dup r> -                         \ does offset fit?
   IF  dup >r ndrop                 \ remove nibbles
       r> r> Tcp !  swap >source  EXIT
   THEN
   r> nibbles!  drop  false
;

Variable Prefix     Prefix off     \ Prefix codes for conditionals to handle e.g. "carry IF" etc.

$FFFF0000 Constant #prefixmask
    $FFFF Constant #checkmask

: ?pairs  ( n1 n2 -- )  swap #checkmask and ?pairs ;

        0 Constant #no-bit
   $10000 Constant #c-bit
   $20000 Constant #s-bit
   $30000 Constant #z-bit
   $40000 Constant #?dup-bit

: noprefix     ( -- )      #no-bit Prefix ! ;

: check#       ( # -- n )  #checkmask and ;

: prefix#      ( # -- n )  #prefixmask and ;

: checknumber  ( n -- # )  check#  Prefix @ or ;

: >prefix  ( # -- )   \ executes NOPREFIX upon EXIT in the definition where >prefix is used
   prefix# Prefix !
   r> [ ' noprefix >body ] Literal rel>abs >r >r
;
: ?strange_prefix ( n -- ) ?dup 0=EXIT dup . abort" strange Prefix" ;

: conditional,  ( -- )  Prefix @  noprefix
   #no-bit   case? IF  T z-branch drop H  EXIT THEN   \ standard Forth consumes the flag
   #z-bit    case? IF  T z-branch      H  EXIT THEN
   #?dup-bit case? IF  T z-branch      H  EXIT THEN
   #c-bit    case? IF  T nc-branch      H  EXIT THEN
   #s-bit    case? IF  T ns-branch      H  EXIT THEN
   ?strange_prefix
;
: z-branch?  ( -- f )  Prefix @ dup #no-bit = swap #?dup-bit = or ;
: ?drop,     ( -- )    z-branch? IF  T drop H  THEN ;
: ?dup,      ( -- )    z-branch? IF  T dup  H  THEN ;

: unconditional,  ( -- )  T branch H ?drop, noprefix ;

$101 Constant #if
$202 Constant #else
$303 Constant #begin
$404 Constant #while
$505 Constant #for
$606 Constant #colon

: (if   ( quan.offset -- sca caddr # )
   >mark    #if checknumber   conditional,
;
: (else ( quan.offset -- sca caddr # )
   >mark  #else checknumber unconditional,
;

Asm NEVER NONE BRA H #codemask Match nop?

: (begin  ( prefixcode -- sca caddr # )
   dup >r #no-bit =
   IF  there 1- t@ nop? Tcp +!       \ optimize a preceeding NOP to become DUP
       T dup <mark drop H            \ handle standard Forth UNTIL (see: UNTIL)
   ELSE  <mark
   THEN
   r> #begin or
;
Asm NEVER PUSH BRA H #codemask Match dup?
#?dup $18 shift #macromask     Match ?dup?

: ?dup_before_z-branch  ( -- )  \ when DUP preceedes IF/UNTIL, no DROP needed
                               \ when ?DUP preceedes IF/UNTIL, only DROP when branch is not taken
   Prefix @ #no-bit = 0=EXIT
   prev@        dup? IF     #z-bit Prefix !  -1 Tcp +!  EXIT THEN
   there 1- t@ ?dup? IF  #?dup-bit Prefix !  -3 Tcp +!  EXIT THEN
;

Asm ALWAYS BOTH BRA H #codemask Match call?  

Target definitions Forth

: carry ( -- )    ?comp #c-bit Prefix ! ;
: sign  ( -- )    ?comp #s-bit Prefix ! ;

: IF    ( -- sca caddr # )  ?comp ?dup_before_z-branch 1 (if ;

: ELSE  ( sca1 caddr1 #1 -- sca1 caddr1 sca2 caddr2 #2 )
   ?comp dup #if ?pairs  >prefix 1 (else
;
\ Due to the variable length branch offset fields ELSE does not compile anything
\ This is handled by THEN instead. ELSE leaves its unique check digit. Therefore,
\ THEN "knows" when it was preceeded by an ELSE.
\
\ THEN tests the check digit: If it was #ELSE,
\ the branch between ELSE ... THEN is compiled first yielding the actual branch
\ offset needed for ELSE. Thereafter, the IF ... ELSE branch is compiled.

: THEN  ( sca caddr # | sca1 caddr1 sca2 caddr2 #2 -- )
   ?comp dup check# #if case? 
   IF  >prefix  ?dup, >resolve ?dup IF  (if  EXIT THEN  ?drop,  EXIT THEN
   #else ?pairs  >prefix
   dup >r  >resolve ?dup IF  rdrop (else  EXIT THEN
   r>   Tcp @ >r   >branch drop 1+ Tcp !
   >resolve ?dup IF  rdrop (if  EXIT THEN   r> Tcp !
;
: BEGIN ( -- sca caddr # )   ?comp #z-bit (begin ;

: WHILE ( sca1 caddr1 #1 -- sca2 caddr2 sca1 caddr1 #2 )
   ?comp #begin ?pairs  T IF H drop 2swap #while
;
: REPEAT ( sca caddr 3 | sca2 caddr2 sca1 caddr1 4 -- )
   ?comp -rot  <resolve T branch H
   dup check# #begin case? IF  drop  EXIT THEN
   #while ?pairs
   >prefix >resolve ?dup IF  (if  EXIT THEN  ?drop,
;
: UNTIL  ( sca caddr # -- )
   ?comp dup #begin ?pairs  ?dup_before_z-branch
   prefix# #no-bit -  Prefix @ #no-bit = and
   IF  Tcp !  >source   #no-bit (begin  EXIT THEN
   <resolve conditional,
;
\ UNTIL re-compiles the loop when a standard Forth UNTIL is encountered.
\ A DUP is inserted before BEGIN and a DROP is inserted after BEGIN to
\ get rid of the UNTIL-flag

False [IF]  \ TOR_reg

: FOR    ( -- sca caddr # )   ?comp T >r H <mark #for ;

Asm NZERO POP BRA Op: (next dont H

: NEXT   ( sca caddr # -- )   ?comp #for ?pairs  <resolve T (next rdrop H ;

: 0=EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp Prefix @  noprefix
   #no-bit case? IF  T  z-exit H  EXIT THEN
   #c-bit  case? IF  T nc-exit H  EXIT THEN
   #s-bit  case? IF  T ns-exit H  EXIT THEN
   ?strange_prefix
;
: ?EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp Prefix @  noprefix
   #no-bit case? IF  T    nz-exit H         EXIT THEN
   #c-bit  case? IF  T cc nc-exit H         EXIT THEN
   #s-bit  case? IF  T sign IF exit THEN H  EXIT THEN
   ?strange_prefix
;

[ELSE]

: FOR    ( -- sca caddr # )  ?comp T 1+ H <mark T 1- >r H #for ;

: NEXT   ( sca caddr # -- )  ?comp #for ?pairs  T r> H <resolve T nz-branch drop H ;

: 0=EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp Prefix @  noprefix
   #no-bit case? IF  T z-exit H   EXIT THEN
   #c-bit  case? IF  T nc-exit H  EXIT THEN
   #s-bit  case? IF  T ns-exit H  EXIT THEN
   ?strange_prefix
;
: ?EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp Prefix @  noprefix
   #no-bit case? IF  T nz-exit H            EXIT THEN
   #c-bit  case? IF  T cc nc-exit H         EXIT THEN
   #s-bit  case? IF  T sign IF exit THEN H  EXIT THEN
   ?strange_prefix
;
: ?CALL  ( <name> -- ) ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp ' >body @ >r   Prefix @ noprefix
   #no-bit case? IF              ?nop,  r> dup <resolve T nz-call H    EXIT THEN
   #c-bit  case? IF              T cc H r> dup <resolve T  nc-call H   EXIT THEN
   #s-bit  case? IF  T sign IF H ?nop,  r> dup <resolve T call THEN H  EXIT THEN
   ?strange_prefix
;
: 0=CALL  ( <name> -- ) ( zero: n -- ) ( sign: n -- n ) ( carry: -- )
   ?comp ' >body @ dup  ?nop, <resolve   Prefix @ noprefix
   #no-bit case? IF  T z-call H   EXIT THEN
   #c-bit  case? IF  T nc-call H  EXIT THEN
   #s-bit  case? IF  T ns-call H  EXIT THEN
   ?strange_prefix
;

[THEN]

\
\ data types in data RAM
\
Target definitions  Forth

: code-origin ( caddr -- ) ?exec Tcp ! | ;
: here        ( -- caddr ) ?exec there ;

: data-origin ( taddr -- ) ?exec Tdp !  ;
: allot       ( n -- )     ?exec Tdp +! ;

: Constant    ( n -- )     Constant immediate
Does>         ( -- n )     @ comp? 0=EXIT  T Literal H ;

: Create      ( -- )       Tdp @ T Constant H ;                \ compiles a data memory address

: Variable    ( -- )       T Create H 1 Tdp +! ;

: Label       ( -- )       there Constant immediate ;          \ returns a code memory address

: [           ( -- )       ?comp postpone [ ; immediate

: ]       ( -- )       ?exec ] ;

: #reset  ( -- addr )       ?exec  0 ;                                \ reset vector address
: #irq    ( -- addr )       ?exec  4 ;                                \ interrupt vector address
: #pause  ( -- addr )       ?exec  8 ;                                \ pause vector address
\
\ Words which are useful to execute interactively during target compilation.
\ To be extended when the need arises
\
: .        ?exec . ;
: .s       ?exec .s ;
: u.       ?exec u. ;
: (              postpone ( ; immediate
: \              postpone \ ; immediate
: decimal  ?exec decimal ;
: hex      ?exec hex ;
: include  ?exec include ;

\
\ setting the search order for Target compilation
\
Forth definitions

Variable Colons  Colons off  \ Builds a linked list of :-definitions for the disassembler

Target definitions Forth

 : :    ( <name> -- # )
        ?exec Create there , ]  #colon     \ leaves #colon during compilation for ";" to recognize
        here Colons @ , Colons !  |
Does>       ( -- )     ?comp ?nop, @ dup <resolve T call H ;

: ;    ( # -- )
   ?comp #colon case? IF  T exit H postpone [ EXIT  THEN  postpone ; ; immediate

 : '    ( <name> -- caddr )   ?exec ' >body @ ;

: [']  ( <name> -- caddr )   ?comp ' >body @ lit, ;

Forth Root definitions

' order Alias order
' words Alias words

: Host    ( -- )
   ['] _NUMBER,     IS NUMBER,                    \ restore "normal" INTERPRET behaviour
   ['] APP-COMPILE, IS COMPILE,
   Only Forth also definitions ;

Target definitions Forth

: Macros  ( -- )  ?exec Host Target definitions ; \ This allows macro definitions during target compilation
                                                  \ switch back to target compilation with TARGET
Forth definitions

: t_NUMBER,      ( d -- )
   DOUBLE? 0= IF DROP THEN  lit, ;

: Target  ( -- )
   ['] t_NUMBER, IS NUMBER,                       \ make INTERPRET compile numbers into the target memory
   ['] execute   IS COMPILE,                      \ execute all words during target compilation even in compile mode
   Only Target also definitions
;
: new      ( -- )                                 \ Initialize cross-compiler for another compilation run
   ." MicroCore cross-compiler " .version ." vom " .date ." by KS" cr
   Memory [ #prog #maxprog min cells ] Literal erase
   Tcp off  Tdp off  Prefix off  Colons off   Sequential off
   Macro off   [ Macros @ ] Literal Macros !   [ Macro-link @ ] Literal Macro-link !
   [ Operators @ ] Literal Operators !
;
: Bin-file   ( <name> -- )                       \ write program memory image to file <name> as binary
   BL word dup c@ 0= abort" file name required"
   count R/W BIN           Create-file abort" File could not be created"  >r
   0 >memory there cells r@ write-file abort" File write error"
   r>                       close-file abort" File close failed"
;
CASEROM [IF]

$0A Constant #lf
$0D Constant #cr
  6 Constant #spaces

Create clause ,"       WHEN 16#0000# => RETURN _00000000_;  "

: >clause  ( offset -- addr )  clause + #spaces + ;

Ascii " $19 >clause c!
Ascii " $22 >clause c!
    #cr $24 >clause c!
    #lf $25 >clause c!

  9 Constant vhdladdr
$1A Constant vhdlcode

: makeaddr ( addr -- string )
   Base @ >r  hex  0 <# # # # # #>  r> Base !
;
: makeopcode ( opcode -- string )  
   Base @ >r  binary  0 <# # # # # # # # # #>  r> Base !
;
: make_whenclause  ( caddr -- )
   dup makeaddr vhdladdr >clause swap cmove
   >memory @ makeopcode vhdlcode >clause swap cmove
;
: writeclause  ( handle -- )
   Clause count rot write-file abort" File write error"
;
Create prolog ," prolog_case.vhd"
Create epilog ," epilog_case.vhd"

: file>here  ( string -- addr len )
   count r/o open-file abort" PROLOG/EPILOG file not present" >r
   here 2000 r@ read-file abort" PROLOG/EPILOG: could not read"
   here swap   r> close-file drop
;
: Vhdl-file  ( <name> -- )
   BL word dup c@ 0= abort" file name required"
   count R/W BIN create-file abort" File could not be created" >r
   prolog file>here r@ write-file abort" PROLOG write error"
   r> 0 there bounds  DO  I make_whenclause  dup writeclause  LOOP  >r
   epilog file>here r@ write-file abort" EPILOG write error"
   r> close-file abort" File close failed" ;

[ELSE] ( CASEROM )

$0A Constant #lf
$0D Constant #cr
  0 Constant #spaces

Create clause ," _00000000_,  "

: >clause  ( offset -- addr )  clause + #spaces + ;

Ascii " $1 >clause c!
Ascii " $A >clause c!
    #cr $C >clause c!
    #lf $D >clause c!

$2 Constant vhdlcode

: makeopcode ( opcode -- string )  
   Base @ >r  binary  0 <# # # # # # # # # #>  r> Base !
;
: make_initstring  ( caddr -- )
   >memory @ makeopcode vhdlcode >clause swap cmove
;
: writeclause  ( handle -- )
   Clause count rot write-file abort" File write error"
;
Create prolog ," prolog_rom.vhd"
Create epilog ," epilog_rom.vhd"

: file>here  ( string -- addr len )
   count r/o open-file abort" PROLOG/EPILOG file not present" >r
   here 2000 r@ read-file abort" PROLOG/EPILOG: could not read"
   here swap   r> close-file drop
;
: Vhdl-file  ( <name> -- )
   BL word dup c@ 0= abort" file name required"
   count R/W BIN create-file abort" File could not be created" >r
   prolog file>here r@ write-file abort" PROLOG write error"
   r> 0 there bounds  DO  I make_initstring  dup writeclause  LOOP  >r
   epilog file>here r@ write-file abort" EPILOG write error"
   r> close-file abort" File close failed" ;

[THEN] ( CASEROM )

